/*-
 * $Id: rr-message.c,v 1.14 2002/05/13 11:52:50 jonas Exp $
 *
 * See the file LICENSE for redistribution information. 
 * If you have not received a copy of the license, please contact CodeFactory
 * by email at info@codefactory.se, or on the web at http://www.codefactory.se/
 * You may also write to: CodeFactory AB, SE-903 47, Ume, Sweden.
 *
 * Copyright (c) 2002 Jonas Borgstrm <jonas@codefactory.se>
 * Copyright (c) 2002 Daniel Lundin   <daniel@codefactory.se>
 * Copyright (c) 2002 CodeFactory AB.  All rights reserved.
 */

#include <librr/rr.h>
#include <string.h>

static GObjectClass *parent_class = NULL;

static void
finalize (GObject *object)
{
	RRMessage *message = (RRMessage *)object;

	rr_frame_aggregator_free (&message->frame_list);
	g_hash_table_destroy (message->mime_headers);
	parent_class->finalize (object);
}


static void
rr_message_init (GObject *object)
{
	RRMessage *msg = (RRMessage *)object;
	msg->msgno = -1;
	msg->aggregate_frames = TRUE;

	msg->mime_headers = g_hash_table_new_full (g_str_hash, g_str_equal,
						   g_free, g_free);
}


static void
rr_message_class_init (GObjectClass *klass)
{
	klass->finalize = finalize;

	parent_class = g_type_class_peek_parent (klass);
}


GType 
rr_message_get_type (void)
{
	static GType rr_type = 0;

	if (!rr_type) {
		static GTypeInfo type_info = {
			sizeof (RRMessageClass),
			NULL,
			NULL,
			(GClassInitFunc) rr_message_class_init,
			NULL,
			NULL,
			sizeof (RRMessage),
			16,
			(GInstanceInitFunc) rr_message_init
		};
		rr_type = g_type_register_static (G_TYPE_OBJECT, "RRMessage", 
						  &type_info, G_TYPE_FLAG_ABSTRACT);
	}
	return rr_type;
}

RRFrame *
rr_message_get_frame (RRMessage *message, gsize max_size)
{
	g_return_val_if_fail (RR_MESSAGE_GET_CLASS (message)->get_frame, NULL);

	return RR_MESSAGE_GET_CLASS (message)->get_frame (message, max_size);
}


gboolean
rr_message_process_frame (RRMessage *message, RRFrame *frame, 
			  GError **error)
{
	g_return_val_if_fail (RR_MESSAGE_GET_CLASS (message)->process_frame,
			      FALSE);

	if (message->aggregate_frames) {
		RRFrame *new_frame;

		new_frame = rr_frame_aggregate (&message->frame_list, frame);

		if (new_frame) {
			gint ret;
			
			ret = RR_MESSAGE_GET_CLASS (message)->
				process_frame (message, frame, error);

			g_object_unref (G_OBJECT (new_frame));
			return ret;
		}
		else
			return TRUE;
	}
	return RR_MESSAGE_GET_CLASS (message)->process_frame (message, frame, 
							      error);
}

void
rr_message_set_channel (RRMessage *message, RRChannel *channel)
{
	g_return_if_fail (RR_IS_MESSAGE (message));
	g_return_if_fail (RR_IS_CHANNEL (channel));

	message->channel = channel;
}

gboolean
rr_message_mime_parse_headers (RRMessage *message, RRFrame *frame)
{
	gchar *body, *value, *buf;
	gchar *header = NULL;
	gchar *tok = NULL;

	g_return_val_if_fail (RR_IS_MESSAGE (message), FALSE);
	g_return_val_if_fail (RR_IS_FRAME (frame), FALSE);

	body = rr_frame_mime_get_body (frame);
	if (NULL == body) {
		return TRUE;
	} 
	
	/* Parse headers */
	for (buf = frame->payload; buf < body; buf++) {

		/* Skip leading spaces */
		if (' ' == *buf || '\r' == *buf || '\n' == *buf)
			continue;

		/* Name */
		if (NULL == tok) {
			tok = buf;
			buf = memchr (buf, ':', body - buf);
			if (NULL == buf) {
				goto err;
			}
			header = g_strndup (tok, buf - tok);
			continue;

		/* Value */
		}  else {
			tok = buf;
			buf = g_strstr_len (buf, body - buf, "\r\n");
			if (NULL == buf) {
				goto err;
			}
			value = g_strndup (tok, buf - tok);
			g_hash_table_insert (message->mime_headers, header, 
					     value);
			g_print ("rr_message_mime_parse_headers: MIME Header "
				 "'%s'=>'%s' found.\n", header, value);
			tok = NULL;
			buf ++;
		}
	}

	return TRUE;
 err:
	g_print ("rr_message_mime_parse_frame: "
		 "Error processing MIME headers.\n");
	return FALSE;
}


void
rr_message_mime_header_set (RRMessage *message, const gchar *header, 
			    const gchar *value)
{
	g_return_if_fail (RR_IS_MESSAGE (message));

	g_hash_table_insert (message->mime_headers, g_strdup (header), 
			     g_strdup (value));
}


gchar *
rr_message_mime_header_get (RRMessage *message, const gchar *header)
{
	g_return_val_if_fail (RR_IS_MESSAGE (message), NULL);

	return g_hash_table_lookup (message->mime_headers, header);
}

/**
 * rr_message_set_aggregate:
 * @message: a #RRMessage
 * @aggregate: %TRUE or %FALSE.
 * 
 * Selects if the frames should be aggregated or not.
 **/
void
rr_message_set_aggregate (RRMessage *message, gboolean aggregate)
{
	g_return_if_fail (RR_IS_MESSAGE (message));

	message->aggregate_frames = aggregate;
}

/**
 * rr_message_get_aggregate:
 * @message: a #RRMessage
 * 
 * Returns %TRUE if the frames are aggregated or %FALSE if they are
 * passed to #rr_message_process_frame directly.
 * 
 * Return value: %TRUE or %FALSE. 
 **/
gboolean
rr_message_get_aggregate (RRMessage *message)
{
	g_return_val_if_fail (RR_IS_MESSAGE (message), FALSE);

	return message->aggregate_frames;
}
